@@ -1,34 +1,37 @@
-# Copyright 2015-2021 Laszlo Attila Toth
-# Distributed under the terms of the Apache License, Version 2.0
+#  Copyright 2015-2022, Laszlo Attila Toth
+#  Distributed under the terms of the Apache License, Version 2.0
 
-import argparse
 import os
 import sys
+import time
 import typing
 
 import click
-from click._unicodefun import _verify_python_env
 
 from dewi_core.appcontext import ApplicationContext
 from dewi_core.command import Command
 from dewi_core.commandregistry import CommandRegistry
 from dewi_core.config.node import Node
-from dewi_core.config_env import EnvConfig, ConfigDirRegistry
+from dewi_core.config_env import ConfigDirRegistry, EnvConfig
 from dewi_core.loader.loader import PluginLoader
-from dewi_core.logger import LogLevel
-from dewi_core.logger import log_debug, create_logger_from_config, LoggerConfig
+from dewi_core.logger import LogLevel, LoggerConfig, create_logger_from_config, log_debug
 from dewi_core.optioncontext import OptionContext
 from dewi_core.utils.exception import print_backtrace
 from dewi_core.utils.levenshtein import get_similar_names_to
+from dewi_core.utils.time import humanize_time
+
+try:
+    from click._unicodefun import _verify_python_env
+
+    def _dummy():
+        pass
 
 
-def _dummy():
+    # the UTF-8 check is buggy, ignore it
+    _verify_python_env.__code__ = _dummy.__code__
+except ImportError:
     pass
 
-
-# the UTF-8 check is buggy, ignore it
-_verify_python_env.__code__ = _dummy.__code__
-
 CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])
 
 
@@ -85,30 +88,39 @@ def _print_exception(ctx: ApplicationContext, exc: BaseException):
     print(f'Exception: {exc} (type: {type(exc).__name__})', file=sys.stderr)
 
 
-class AppGroup(AliasedGroup):
-    def main(self, *args, **kwargs):
+def _app_main(parent_cls):
+    def app_main(self, *args, **kwargs):
+        start = end = 0
+
+        def print_run_time():
+            nonlocal end
+            end = time.time()
+            if os.environ.get('DEWI_SHOW_RUN_TIME', '0') == '1':
+                diff = int((end - start) * 100) / 100
+                print(f'Runtime: {humanize_time(end - start, format=True)} ({diff:2} seconds)')
+
+        start = time.time()
         try:
-            super().main(*args, **kwargs)
+            parent_cls.main(self, *args, **kwargs)
         except SystemExit:
+            print_run_time()
             _wait_for_termination_if_needed(self.callback.app_context)
             raise
         except BaseException as exc:
+            print_run_time()
             _print_exception(self.callback.app_context, exc)
-            _wait_for_termination_if_needed(self.callback.app_context)
-            sys.exit(1)
+        _wait_for_termination_if_needed(self.callback.app_context)
+        sys.exit(1)
+
+    return app_main
+
+
+class AppGroup(AliasedGroup):
+    main = _app_main(AliasedGroup)
 
 
 class AppCommand(click.Command):
-    def main(self, *args, **kwargs):
-        try:
-            super().main(*args, **kwargs)
-        except SystemExit:
-            _wait_for_termination_if_needed(self.callback.app_context)
-            raise
-        except BaseException as exc:
-            _print_exception(self.callback.app_context, exc)
-            _wait_for_termination_if_needed(self.callback.app_context)
-            sys.exit(1)
+    main = _app_main(click.Command)
 
 
 def get_invoked_subommand(ctx: click.Context):
@@ -118,9 +130,9 @@ def get_invoked_subommand(ctx: click.Context):
         return ctx.invoked_subcommand
 
 
-def get_command_from_plugin_ns(plugin_name: str, ns: argparse.Namespace) -> typing.Type[Command]:
+def get_command_from_plugin_appcontext(plugin_name: str, ctx: ApplicationContext) -> typing.Type[Command]:
     """
-    Returns the command  class specified in the ns namespace via its running_command_
+    Returns the command  class specified in the ctx ApplicationContext via its running_command_
     and running_subcommands_ members.
 
     The ns comes from an already parsed namespace, especially via
@@ -128,34 +140,32 @@ def get_command_from_plugin_ns(plugin_name: str, ns: argparse.Namespace) -> typi
 
     The command must exist.
 
-    :param plugin_name: The command plugin or app plugin containing the command referred by ns.running_command_
-    :param ns: the prepared Namespace object
-    :return: the command class based on ns.
+    :param plugin_name: The command plugin or app plugin containing the command referrered in ctx
+    :param ctx: the prepared ApplicationContext object
+    :return: the command class based on ctx.
     """
     command_registry = CommandRegistry()
-    cfg_dir_registry = ConfigDirRegistry(EnvConfig(ns.environment_))
-    cfg_dir_registry.register_config_directories(ns.config_directories_)
+    cfg_dir_registry = ConfigDirRegistry(EnvConfig(ctx.environment))
+    cfg_dir_registry.register_config_directories(ctx.config_directories)
     loader = PluginLoader(command_registry, cfg_dir_registry)
     loader.load([plugin_name])
     cfg_dir_registry.load_env()
-    cmd_class = command_registry.get_command_class_descriptor(ns.running_command_).get_class()
+    cmd_class = command_registry.get_command_class_descriptor(ctx.command_names.command).get_class()
 
-    if 'running_subcommands_' in ns and ns.running_subcommands_:
-        for sub_cmd_name in ns.running_subcommands_:
+    if ctx.command_names.subcommands:
+        for sub_cmd_name in ctx.command_names.subcommands:
             for cc in cmd_class.subcommand_classes:
                 if cc.name == sub_cmd_name:
                     cmd_class = cc
 
-    ns.cmd_class_ = cmd_class
-
     return cmd_class
 
 
-def run_command_from_plugin_ns(plugin_name: str, ns: argparse.Namespace) -> typing.Optional[int]:
+def run_command_from_appcontext(plugin_name: str, ctx: ApplicationContext) -> typing.Optional[int]:
     """
-    Runs the command returned by get_command_from_plugin_ns(). See that function.
+    Runs the command returned by get_command_from_plugin_appcontext(). See that function.
     """
-    return get_command_from_plugin_ns(plugin_name, ns)().run(ns)
+    return get_command_from_plugin_appcontext(plugin_name, ctx)().run(ctx)
 
 
 def _list_commands(prog_name: str, command_registry: CommandRegistry, *, all_commands: bool = False):
@@ -223,12 +233,14 @@ class Application:
     def __init__(self, program_name: str,
                  command_class: typing.Optional[typing.Type[Command]] = None,
                  *,
+                 description: typing.Optional[str] = None,
                  enable_short_debug_option: bool = False,
                  enable_env_options: bool = True,
                  version: typing.Optional[str] = None
                  ):
         self._program_name = program_name
         self._command_class: typing.Optional[typing.Type[Command]] = command_class
+        self._description = description
         self._enable_short_debug_option = enable_short_debug_option if command_class is not None else True
         self._enable_env_options = enable_env_options if command_class is not None else True
         self._version = version
@@ -302,17 +314,26 @@ class Application:
 
         app_run.app_context = app_context
 
-        app_click_helper = OptionContext()
-        app_click_helper.register_args(self._register_app_args)
+        app_opt_ctx = OptionContext()
+        app_opt_ctx.register_args(self._register_app_args)
         if single_command_mode:
-            app_click_helper.register_args(self._command_class.register_arguments)
+            app_opt_ctx.register_args(self._command_class.register_arguments)
 
-        app_run = app_click_helper.add_args_to_func(app_run)
+        try:
+            app_run = app_opt_ctx.add_args_to_func(app_run)
+        except Exception as e:
+            if single_command_mode:
+                suffix = f'for command `{self._command_class.name}`'
+            else:
+                suffix = 'for the application'
+            print(f'Error occurred while registering args {suffix}: {e}')
+            raise
 
         if single_command_mode:
             has_classes = len(self._command_class.subcommand_classes) > 0
             app_run = (click.group if has_classes else click.command)(
-                self._program_name, help=self._command_class.description,
+                self._program_name,
+                help=self._description or self._command_class.description,
                 context_settings=CONTEXT_SETTINGS,
                 cls=(AppGroup if has_classes else AppCommand))(app_run)
             self._register_subcommands(self._command_class.subcommand_classes, app_run, app_context)
@@ -320,7 +341,10 @@ class Application:
             self._command_registry.register_class(_ListAllCommand)
             self._command_registry.register_class(_ListCommand)
 
-            app_run = click.group(self._program_name, cls=AppGroup, context_settings=CONTEXT_SETTINGS)(app_run)
+            app_run = click.group(self._program_name, cls=AppGroup,
+                                  help=self._description,
+                                  context_settings=CONTEXT_SETTINGS,
+                                  )(app_run)
             self._register_subcommands(self._command_registry._command_classes, app_run, app_context)
 
         return app_run(args, self._program_name)
@@ -349,13 +373,17 @@ class Application:
 
                 return r
 
-            cmd_helper = OptionContext()
-            cmd_helper.register_args(command_class.register_arguments)
-            cls = AliasedGroup if command_class.subcommand_classes else click.core.Command
-            cmd_run = parent_run_cmd.command(name=command_class.name, help=command_class.description, cls=cls,
-                                             context_settings=CONTEXT_SETTINGS)(
-                cmd_helper.add_args_to_func(wrapper()))
-            self._register_subcommands(command_class.subcommand_classes, cmd_run, app_context)
+            try:
+                cmd_opt_ctx = OptionContext()
+                cmd_opt_ctx.register_args(command_class.register_arguments)
+                cls = AliasedGroup if command_class.subcommand_classes else click.core.Command
+                cmd_run = parent_run_cmd.command(name=command_class.name, help=command_class.description, cls=cls,
+                                                 context_settings=CONTEXT_SETTINGS)(
+                    cmd_opt_ctx.add_args_to_func(wrapper()))
+                self._register_subcommands(command_class.subcommand_classes, cmd_run, app_context)
+            except Exception as e:
+                print(f'Error occurred while registering args for command `{command_class.name}`: {e}')
+                raise
 
     def _register_app_args(self, h: OptionContext):
         if self._version:
